;FST4BCN    G4JNT  Mar 2021
;Completely new interrupt routine and frequency calculation
;  derived from WSPRbcn_9852_Rand.   All routines left in, even if not called

	LIST P = 16F628A
	INCLUDE "P16F628A.INC"

    __config    0x3F61      ;LV Programming off , OSC XT Mode
    __idlocs    0x9023      ;
  errorlevel 1, -207  ;Suppress msgs. and warnings about labels after col 1.



    #define     SDIO        PORTA, 0
    #define     SCLK        PORTA, 1
    #define     IOUD        PORTA, 2
    #define     TXENABLE    PORTA, 3	

    #define     PPS         PORTB, 0
    #define     GPSIN       PORTB, 1
    #define     GPSOUT      PORTB, 2
    #define     STROBE      PORTB, 3 		;Used for testing interrupt	
    #define     LEDGrn      PORTB, 4
    #define     LEDRed      PORTB, 5
 
    #define     TESTLINK    PORTB, 7

    #define	StartFlag	Flags, 0
    #define     TestModeFlag    Flags, 1
    #define     GPSOk           Flags, 2
    #define     HexFlag         Flags, 6
    #define     Hopping		Flags, 7


	IGNORETESTLINK	=	0	;Used when programming to avoid programmer pulling it down
	STARTSECOND	=	0		;allows sequence to run late (or early if minutes fixed)
	FSTMODE		=	3		;0 = 15, 1 = 30 ... 7 = 1800

	NSYMBS		=	d'160'	
  	NMEAPol 	= 	0      	;0 = TTL levels, 1 = RS232
  	CWSPEED     	=   	d'50'   ;CW Dot length, ms.    Needed for small delays, so don't set to 0
  	ALWAYSON	=	0	;No on off switching
		
	GPSBAUDDEL	=	d'40'	;6.N + 18 = Fc / Baud   or   N ~ (Fxtal / Baud)/24 - 3   (9600 == 40,  4800 == 84)




	MODEDIV15	=	d'293'	
	MODEDIV30	=	d'684'
	MODEDIV60	=	d'1581'
	MODEDIV120	=	d'3336'
	MODEDIV300	=	d'8749'
	MODEDIV900	=	d'27084'
	MODEDIV1800	=	d'54688'
	MODEDIVTEST	=	d'2441'		;0.5s symbols

	#include	"FST4Bcn.inc"		;Contains all frequecy and details specific to frequency and clock



 org     0
     goto    StartUp

;------------------ Interrupt routine, get here 4 cycles after TMR overflow --
Ints					;Here at 64 * symbol rate

;        movwf   Wreg        	;Save W
;	swapf   STATUS , W      ; nb.  movwf does not affect any STATUS bits
;	movwf   SaveSts         ;Save STATUS, without affecting it

	nop
	nop
	nop

	nop
	bsf     IOUD
	nop
	nop
	bcf     IOUD


	movf	IptCountHi, W
        movwf   TMR1H
	movf	IptCountLo, W
	movwf	TMR1L	

   	bcf     PIR1 , TMR1IF   	;reset the interrupt flag
 ;Below here is not time-critical, but must not overrun

	movf	SampCount, W	;Test for 0
	btfss	STATUS, Z
	goto	NoNewSymbol

	btfsc   SymbCount, 0    ;Flash Red LED at half symbol rate
        bcf     LEDRed		;  On for Odd values
        btfss   SymbCount, 0
        bsf     LEDRed

	movf	ThisSymbol, W	;Shift all symbols back one, and get new NextSymbol
	movwf	LastSymbol	;Make a copy of the symbol before updating it
	movf	NextSymbol, W	;First Pass, ThisSymbol has been pre-loaded
	movwf	ThisSymbol

	movf	SymbCount, W	
	addlw	1		;  thereafter is it copied from NexSymbol after lookup
	call	GetSymbol
	movwf	NextSymbol


	incf	SymbCount	

	movlw   NSYMBS + 1	;Easiest way to make sure final one is actually sent
	subwf   SymbCount, W    ;Test for a full message   W = Count - Threshold, When +ve/0 C=1 do action
	btfsc   STATUS, C
	bcf	INTCON, PEIE    ;Inhibit interrupt after final symbol
;..................
NoNewSymbol	;Here at sample rate with values in Last/This/NextSymbol  and count of 0 - 63. 0 is start of new symbol
				
		;This goes 32 to 64 and back to 32
		;Next goes from 32 to 63 starting from count of 32

	movf	SampCount, W	;Pointer for LastSymbol goes from 32 to 0 then Freezes after SymbCount = 33
	sublw	d'32'
	movwf	Temp		;Temp = 32 - SampCount
	btfsc	Temp, 7		; if negative, freeze at zero
	clrf	Temp
	movf	Temp, W
	call	PTable
	movwf	B0		;B0 contains Weight of LastSample

	movf	LastSymbol, W
	movwf	A0
	clrf	Q0
	clrf	Q1
	call	ShortMultiplyAdd	;First component stored
				;~~~~~~~~~~~~~				
	movf	SampCount, W	;Pointer for ThisSymbol goes from 32 to 64 (at 32) and back to 32
	addlw	d'32'
	movwf	Temp		;W / Temp = SampCount + 32

	btfsc	Temp, 6		;Test for value >= 64
	sublw	d'128'		;if >=64, W = 128 - SampCount - 32, else W = SampCount + 32

	call	PTable
	movwf	B0		;B0 contains Weight of LastSample
	movf	ThisSymbol, W
	movwf	A0
	call	ShortMultiplyAdd	;Primary component stored	
				;~~~~~~~~~~~~~~~~~
	movlw	d'32'		;Pointer for NextSymbol goes from 0 to 31, starting at SymbCount = 32
	subwf	SampCount, W	
	movwf	Temp		;Temp = SampCount - 32
	btfsc	Temp, 7		; if negative, freeze at zero
	clrf	Temp
	movf	Temp, W
	call	PTable
	movwf	B0		;B0 contains Weight of LastSample
	
	movf	NextSymbol, W
	movwf	A0
	call	ShortMultiplyAdd	;Final component stored

	incf	SampCount, W	;Counts 0 - 63 
	andlw	0x3F
	movwf	SampCount

	movf	Q0, W
	movwf	A0

	movf	Q1, W
	movwf	A1		;A1/0  now contain fraction of DDS Freq constants
				; value for this sample as a number from 0 - [3 * 255]
	movf	ModeInUse, W
	movwf	Temp
	rlf	Temp
	rlf	Temp, W
	andlw	0x1C
	movwf	Temp


	movlw	LOW(FreqConsts) + 3
	addwf	Temp, W
	call	GetEE
	movwf	B0

	movlw	LOW(FreqConsts) + 2
	addwf	Temp, W
	call	GetEE
	movwf	B1

	movlw	LOW(FreqConsts) + 1
	addwf	Temp, W
	call	GetEE
	movwf	B2

	movlw	LOW(FreqConsts)
	addwf	Temp, W
	call	GetEE
	movwf	B3

	call	Multiply24x16

	call    GenToneData	;Adds result and puts freq code into D0/5
	movlw   2
	movwf   DDSAddr
	call    DataTo9852      ;Send Freq data, updated on next int	


OutInterrupt
;	swapf   SaveSts , W     ;Restore STATUS
;	movwf   STATUS
;	swapf   Wreg            ;Restore W
;	swapf   Wreg , W        ;Can't use movf as it  affects Z bit

    retfie
;=====================================
StartUp 
	clrf    PORTA       ;Special setup to disable comparator on 'F628
	movlw   7
	movwf   CMCON
	
	bsf     STATUS,RP0      ;ram page 1

	movlw   b'00000001'     ;Pull-ups
	movwf   OPTION_REG
	movlw   b'00000000'     ;
	movwf   PORTA           ;
	movlw   b'10001011'     ;B0 PPS, B1 GPSIN, B2 GPSOUT B3 RXD 
	movwf   PORTB           ;B6 ext key output, B7 test mode input

	bsf	PIE1, TMR1IE
	bcf     STATUS,RP0      ;ram page 0

	movlw	HIGH(PTable)		;All tables in the same page, so use any one of them here
	movwf	PCLATH
	clrf    PORTB
	clrf    PORTA

	movlw	0		;
	movwf	SymbCount
	movwf	SampCount

	movlw	b'00110101'	;Prescale/8 Osc off, no sync, Internal clock, Timer 1 Disabled for now
	movwf	T1CON

	clrf    Flags
	movlw   0xFF
	movwf   PNCounter       ;Preloaded so we get 0 at first iteration

	call    InitGPS

	call	SetUpDDS
	movlw	0
	call	OnOff
	movlw	d'40'
	call	ProgDelay

	btfss	TESTLINK	;Set at turn on reads duty cycle/hop status from RC
	call	ReadSettings

	call	ReadCR
	sublw	7		;So pot works in opposite direction
	movwf	ModeInUse

	bcf	TXENABLE
	addlw	"1"
	call	SendCWChar
	movlw	d'200'
	call	ProgDelay
	movf	ModeInUse, W
	addlw	"1"
	call	SendCWChar

	movlw	d'200'
	call	ProgDelay
	call	ShowDC		;Shows / plays Duty cycle and hop status

	call    InitGPS		;Do it twice, just in case

WaitGPSFirstLock            ;
                           
	btfss   TESTLINK        	
	call    TestMode	

	call    GetGPSData      
	btfss   GPSOk
	goto    WaitGPSFirstLock

	bcf     LEDRed

	movlw	0x12		;Randomise starting frequency
	movwf	V1
	movf	Seconds, W
	movwf	V2	
	movf	Minutes, w
	movwf	V3
	clrf	V4
	call	SpreadLoop
	call	SpreadLoop
	call	SpreadLoop
SkipGPS

	movf	ModeInUse, W	
	movwf	Temp
	bcf	STATUS, C
	rlf	Temp, W
	call	ModeDivTable	;Set interupt timer reload value
	movwf	IptCountHi

	bsf	STATUS, C
	rlf	Temp, W
	call	ModeDivTable
	movwf	IptCountLo


	bcf	PIR1, TMR1IF	;In case it has been set before GIE is enabled
	bcf	INTCON, PEIE	
	bsf	INTCON, GIE	;  and set it running
;...........................

MainLoop

	btfss   TESTLINK	
	call    TestMode	

	call    GetGPSData      ;Sets Minutes, Seconds and RxStatus
	call	TestStart
	btfss	StartFlag
    	goto    CarryOn 
;................
StartSend

     	call    PNSeq       ;Tx with set Duty cycle in randomised slots
    	movwf   CheckDigit  
    	iorlw   0           ;Returns with W = 0 to Tx
    	btfss   STATUS, Z
    	goto    NoTxValue   ;Gives special flash pattern (and delays past the next second)

	call	SpreadLoop

	movf	V1, W
	andlw	0x7F
	movwf	RandOffset	;Value from 0 - 127

    	call    SetUpDDS    ;Comes on at full power
    	movlw   0           ;But turn off RF for now unless ALWAYSOn = 1
    	call    OnOff

    	bsf     TXENABLE    ;Enable Tx up to 1s earlier to allow for relays
    	call    WaitPPS     ;Start on the minute boundary + 1s

    	movlw   1            ;Turn on RF
    	call    OnOff

    	clrf    SymbCount    ;Re-Initialise registers
	clrf	SampCount
	clrf	ThisSymbol	;Ensues LastSymbol is set to zero for the start

	movlw	0		;Pre-load NextSymbol which is shifted to ThisSymbol
	call	GetSymbol	;  on first pass into interrupt routine
	movwf	NextSymbol

	bcf	PIR1, TMR1IF
	bsf	INTCON, PEIE	    ;Enable the interrupt, starts FST4 sequence
	bsf	T1CON, TMR1ON
FST4Loop
    	btfsc   INTCON, PEIE ;Sit in here waiting for completion   
    	goto    FST4Loop

	bcf	T1CON, TMR1ON

	call	SetUpDDS	;Restores tone zero
	

    	movlw   0
    	call    OnOff
    	movlw	d'50'
	call	ProgDelay
	
    	movlw   LOW(CWMsg)
    	movwf   CWMsgCnt
    	call    SendCWMsg

    	bcf     TXENABLE
    	call    CWDelay    
CarryOn
    	goto    MainLoop
;.............................
NoTxValue  
    movlw   4
    movwf   Counter     ;Unique flash pattern to show Random value
NoTxLoop                ;  when transmit sequence has NOT been triggered
    bcf     LEDGrn      ; Flashes R+G 
    bsf     LEDRed
    call    CWDelay
    bsf     LEDGrn
    bcf     LEDRed
    call    CWDelay
    bcf     LEDGrn
    call    CWDelay
    call    CWDelay
    call    CWDelay
    call    CWDelay
    call    CWDelay

    decfsz  Counter
    goto    NoTxLoop

    goto    MainLoop
;===========================================
TestMode                    ;TESTLINK Input grounded
 if IGNORETESTLINK == 1
 	return
 endif		


	call    SetUpDDS        ;  Stays here until link is removed
	bsf     LEDRed
	bcf     LEDGrn
	bsf     TXENABLE
TestModeLoop
	bsf	LEDRed
	bcf	LEDGrn
	movlw	d'10'
	call	ProgDelay
	bcf	LEDRed
	bsf	LEDGrn
	movlw	d'10'
	call	ProgDelay

	btfss   TESTLINK
	goto    TestModeLoop

	bcf     TXENABLE
	call    CWDelay
	movlw   0
	call    OnOff

	bcf     LEDGrn

	return

;------------------------------------------------
GetSymbol		;Enter with Symbol Value. Lookup quad and get correct bit pair
	movwf	Temp
	andlw   0x03            ;Stores position in the QuadSymbol
        movwf   Counter

	rrf	Temp	
	rrf	Temp, W
	andlw	0x3F		;W = SymbCount / 4,
	addlw	LOW(FST4MsgData)
	call	GetEE
	movwf	QuadSymbol	
 
        incf    Counter    	;want 1,2,3 or 4 passes through alignment loop
PosDataLoop
        rlf     QuadSymbol      ;Data items are left justified in each byte
        rlf     QuadSymbol      ;rotate 2, 4, or 6 times though C
        decfsz  Counter
        goto    PosDataLoop

        rlf     QuadSymbol, W   ;One final shift to position at LSBs
        andlw   0x03            ; total of 3,5,7, or 9 left shifts/rotates

	return			;Return with symbol value in W
;------------------------------------------------
GenToneData     
        movlw   LOW(FreqData) + 5    ;Get base carrier frequency
	call    GetEE
	movwf   D0
        movlw   LOW(FreqData) + 4
	call    GetEE
	movwf   D1
        movlw   LOW(FreqData) + 3
	call    GetEE
	movwf   D2
        movlw   LOW(FreqData) + 2
	call    GetEE
	movwf   D3
        movlw   LOW(FreqData) + 1
	call    GetEE
	movwf   D4
        movlw   LOW(FreqData) 
	call    GetEE
	movwf   D5              

 	movf	Q0, W			;Add in the result from the multiplication
	addwf	D0

	movf	Q1, W
	btfsc	STATUS, C
	incfsz	Q1, W		
	addwf	D1

	movf	Q2, W
	btfsc	STATUS, C
	incfsz	Q2, W
	addwf	D2		

	movf	Q3, W
	btfsc	STATUS, C
	incfsz	Q3, W
	addwf	D3

	movf	Q4, W
	btfsc	STATUS, C
	incfsz	Q4, W
	addwf	D4

	movlw	0
	btfsc	STATUS, C
	movlw	1
	addwf	D5


	movf	RandOffset, W		;Value from 0 - 127
	addlw	3			;To get away from bottom of band
	movwf	A0
	clrf	A1

        movlw   LOW(RandJump) + 1		;Get Random jump increment
	call    GetEE
	movwf   B2
        movlw   LOW(RandJump) + 2
	call    GetEE
	movwf   B1
        movlw   LOW(RandJump) + 3
	call    GetEE
	movwf   B0

	call	Multiply24x16

 	movf	Q0, W			;Add in the result from the randomisatino
	addwf	D0

	movf	Q1, W
	btfsc	STATUS, C
	incfsz	Q1, W		
	addwf	D1

	movf	Q2, W
	btfsc	STATUS, C
	incfsz	Q2, W
	addwf	D2		

	movf	Q3, W
	btfsc	STATUS, C
	incfsz	Q3, W
	addwf	D3

	movf	Q4, W
	btfsc	STATUS, C
	incfsz	Q4, W
	addwf	D4

	movlw	0
	btfsc	STATUS, C
	movlw	1
	addwf	D5

	


	return
;--------------------------------------
ShortMultiplyAdd		;Multiply Byte in B * two bits in A,  Add to Q
	btfss	A0, 0
	goto	DoTwo

	movf	B0, W
	addwf	Q0
	btfsc	STATUS, C
	incf	Q1
DoTwo
	btfss	A0, 1
	return

	bcf	STATUS, C
	rlf	B0, W
				
	btfsc	STATUS, C	;First deal with MSB shifted into C
	incf	Q1
	
	addwf	Q0
	btfsc	STATUS, C
	incf	Q1

	return
;------------------------------
Multiply24x16      	; Up to ~ 380 clocks max
	clrf    Q0        ;Input A0/A1 and B0/1/2, output Q0/1/2/3/4
	clrf    Q1        ;A destroyed,  B kept
	clrf    Q2        
	clrf    Q3  
	clrf	Q4     

	movlw   d'16'       ;A and B 24 bit
	movwf   Counter
MultLoop            
	bcf     STATUS, C
	rrf	A1
	rrf     A0         ;Working bit into carry
	btfss   STATUS, C   ;  into Q depending on working bit of A
	goto    NoMult


	movf    B0, W       ;get shifted B0 value
	addwf   Q2          ; and add into accumulator if 1 in 
			; appropriate bit in Avalue
	movf    B1, W
	btfsc   STATUS, C   
	incfsz  B1, W
	addwf   Q3	

	movf    B2, W
	btfsc   STATUS, C   
	incfsz  B2, W
	addwf   Q4	


NoMult                  ;Need to rotate in carry from final addition
	rrf     Q4        ;C already zero if we've jumped in here
	rrf     Q3
	rrf     Q2
	rrf     Q1
	rrf	Q0
	decfsz  Counter
	goto    MultLoop

	return              ;Q4/3/2/1/0 = A * B     B kept, A destroyed
;------------------------------------------------------------------
WaitPPS                 ;In case PPS is still high
    btfsc   PPS         ; look for a low to high transition
    goto    WaitPPS

WaitRefPPS              ;Now wait for the rising edge itself
    btfss   PPS        
    goto    WaitRefPPS 

    return
;----------------------
GetEE         ;different from ;F84   EE registers all in bank 1 now 
	bsf     STATUS,RP0      ;ram page 1
	movwf   EEADR
	bsf     EECON1 , RD
	movf    EEDATA , W
	bcf     STATUS , RP0    ;ram page 0
	return
;-----------------------
StoreEE           ;Enter with W = EE data,  Whi = EE address
    bsf     STATUS, RP0
    movwf   EEDATA
    bcf     STATUS, RP0
    movf    Whi, W
    bsf     STATUS, RP0
    movwf   EEADR    

    bsf     STATUS , RP0    ;All EE registers in Bank 1

    bsf     EECON1 , WREN   ;Enable EEprom writing
    movlw   0x55
    movwf   EECON2
    movlw   0xAA
    movwf   EECON2
    bsf     EECON1 , WR

    bcf     STATUS, RP0     ;PIR1 register is in bank 0
EEwaitW
   
    btfss   PIR1 , EEIF
    goto    EEwaitW
    bcf     PIR1 , EEIF       ;has to be cleared manually

    bsf     STATUS, RP0
    bcf     EECON1 , WR
    bcf     STATUS , RP0
    return
;-------------------------

SetUpDDS
    movlw   7           ;Control register
    movwf   DDSAddr
    movlw   0x10
    movwf   D3
    movlw   LOW(ControlPLL)      ;PLL factor & range bit
    call    GetEE

    movwf   D2
    movlw   0x00        ;LSB of this is external update
    movwf   D1         
    movlw   0x60        ;OSK Enabled, bypass InV Sinc.
    movwf   D0
    call    DataTo9852

    movlw   2
    movwf   DDSAddr
    movlw   LOW(FreqData) + 5           ;Set CW carrier frequency
    call    GetEE
    movwf   D0
    movlw   LOW(FreqData) + 4
    call    GetEE
    movwf   D1
    movlw   LOW(FreqData) + 3
    call    GetEE
    movwf   D2
    movlw   LOW(FreqData) + 2
    call    GetEE
    movwf   D3
    movlw   LOW(FreqData) + 1
    call    GetEE
    movwf   D4
    movlw   LOW(FreqData)
    call    GetEE
    movwf   D5
    call    DataTo9852

    movlw   8           ;Turn amplitude full on immediately
    movwf   DDSAddr
    movlw   0xFF
    movwf   D0
    movlw   0x0F
    movwf   D1
    call    DataTo9852


	bsf     IOUD
	nop
	nop
	bcf     IOUD

	return
;--------------------------
OnOff			;Raised cosine ramping
  if  ALWAYSON == 1

	movlw	0x0F
	movwf	D1
	movlw	0xFF
	movwf	D0
	movlw   8             
	movwf   DDSAddr
	call    DataTo9852
	bsf     IOUD
	nop
	bcf     IOUD

	return

  else
	
  	iorlw   0       ;W = 0 output off,   W = 1, Output level = max
    	btfsc   STATUS, Z       ;Each amplitude update takes approx 320 clocks
	goto    RampOff         ;  == 130us.  So 8 of these = 1ms
                            
	movlw   d'16'
	movwf   SoftCount       
SoftOnLoop
    	movf    SoftCount, W
	sublw	d'16'		;Convert count of 16 - 1 (skips on final) to 0 - 15 
	call	RampTable
	movwf	Amplitude	;Returns 0 - 255

	swapf	Amplitude, W
	andlw	0x0F		;MS Nibble of Amplitude
	movwf	D1		

	swapf	Amplitude, W	
	andlw	0xF0
	iorlw	0x0F		;Amplitude enters with 000, first pass forces to 00F, after 16 passes through, FFF	
	movwf	D0

	movlw   8               ;16 passes though here ramp the values:0x00F,0xaaF,0xbbF, ... 0xFFF
	movwf   DDSAddr
	call    DataTo9852
	bsf     IOUD
	nop
	bcf     IOUD
    	
    decfsz  SoftCount
    goto    SoftOnLoop

  endif

    return
;............
RampOff
    movlw   d'16'
    movwf   SoftCount       
SoftOffLoop
    	movf    SoftCount, W
	addlw	0xFF		;Convert count of 16 - 1 (skips on final) to 15 to 0 
	call	RampTable
	movwf	Amplitude	;Returns 0 - 255

	swapf	Amplitude, W
	andlw	0x0F		;MS Nibble of Amplitude
	movwf	D1		

	swapf	Amplitude, W	
	andlw	0xF0
	movwf	D0		;Amplitude enters with FFF, first pass forces to FF0, after 16 passes through, 000	

    	movlw   8               ;16 passes though here ramp the values:0xFF0,0xbb0,0xaa0, ... 0x000
	movwf   DDSAddr
	call    DataTo9852
	bsf     IOUD
	nop
	bcf     IOUD
  	
    	decfsz  SoftCount
    	goto    SoftOffLoop

    	return

;--------------------------
Send9852Byte          ;86 clocks incl call and return
	movwf   Temp
	movlw   8
	movwf   Counter
DDSSendLoop
    rlf     Temp          
	btfsc   STATUS, C
	bsf     SDIO
	btfss   STATUS, C
	bcf     SDIO
	bsf     SCLK
	bcf     SCLK

	decfsz  Counter
	goto    DDSSendLoop

	return
;-------------------------
LongDelay           ;Approx 1 second delay
	movlw   d'250'
ProgDelay
	movwf   DelCount2
LongDelLoop
	call    Delay1ms
	call    Delay1ms
	call    Delay1ms
	call    Delay1ms
	decfsz  DelCount2
	goto    LongDelLoop
	return
;-------------------------
Delay1ms        ;5N+5 =  1250 clocks, * 2 = 2500 clocks = 1ms
	movlw   d'249'
	movwf   DelCount
Del1msLoop
	nop
	nop
	decfsz  DelCount
	goto    Del1msLoop
	movlw   d'249'
	movwf   DelCount
Del1ms2Loop
	nop
	nop
	decfsz  DelCount
	goto    Del1ms2Loop
	nop
	nop
	nop
	nop
	return
;-------------------------
GetGPSData          ;Monitors GPSIN line for $GPRMC   (may have to change to G[x]RMC for later nav receivers)
	nop
	bsf     LEDGrn      ;Flash Green LED Until header is received.  If LEDRed
	btfss   GPSOk       ; is enabled, this gives a unique flash pattern
	bcf     LEDGrn      ;Flash in antiphase if GPS is not locked up
	call    PushStack
	call    GetGPSByte
	movwf   Stack0		;String starts    $GPRMC,HHMMSS.ss   (decimal seconds optional)
				;Stack position - CBA9876543210
	movf    StackC , W      
	sublw   "$"        
	btfss   STATUS , Z
	goto    GetGPSData
	movf    StackB , W     
	sublw   "G"
	btfss   STATUS , Z
	goto    GetGPSData
	movf    StackA , W      
	sublw   "P"
	btfss   STATUS , Z
	goto    GetGPSData
	movf    Stack9 , W      
	sublw   "R"
	btfss   STATUS , Z
	goto    GetGPSData
	movf    Stack8 , W      
	sublw   "M"
	btfss   STATUS , Z
	goto    GetGPSData
	movf    Stack7 , W      
	sublw   "C"
	btfss   STATUS , Z
	goto    GetGPSData
	movf    Stack6 , W      ;May be overkill checking all these
	sublw   ","		;   but... whatever....
	btfss   STATUS , Z
	goto    GetGPSData

	movf	Stack2, W	;String starts    $GPRMC,HHMMSS.ss   (decimal seconds optional)
	movwf	BCDLo		;Stack position - CBA9876543210
	movf	Stack3, W
	movwf	BCDHi
	call	BCDtoBIN
	movwf	Minutes

	movf	Stack0, W
	movwf	BCDLo
	movf	Stack1, W
	movwf	BCDHi
	call	BCDtoBIN
	movwf	Seconds

	bcf     LEDGrn      	;Green off means time data has been decoded
	btfss   GPSOk		;Flash in antiphase if GPS is not locked up
	bsf     LEDGrn      	

				;Now continue looking for the comma delimiting the status value
NextComma
	nop
	call    PushStack
	call    GetGPSByte	;After arbitrary seconds formatting...
	movwf	Stack0		;                 xxxx,A,    (or V) 
				;Stack position - 6543210
	sublw	","
	btfss	STATUS, Z
	goto	NextComma
	movf	Stack2, W
	sublw	","
	btfss	STATUS, Z
	goto	NextComma	;
	
	movf	Stack1, W
	sublw	"A"
	bcf	GPSOk
	btfsc	STATUS, Z
	bsf	GPSOk

	return
;----------------------------
BCDtoBIN                ;Converts pairs of ASCII BCD characters to
	movf    BCDHi , W   ;binary in W
	andlw   b'00001111' ;mask from ASCII to digit  = [BCDHi]
	movwf   Temp
	bcf     STATUS , C
	rlf     Temp
	rlf     Temp        ;Temp = 4 * [BCDHi]
	movlw   b'00001111'
	andwf   BCDHi , W   ;W = [BCDHi]
	addwf   Temp        ;Temp = 5 * [BCDHi]
	bcf     STATUS , C
	rlf     Temp
	movf    BCDLo , W
	andlw   b'00001111'
	addwf   Temp        ;Temp = 10 * BCDHi + BCDLo
	movf    Temp , W
	return
;-------------------------------
DoTime                  ;Extract  MMSS from ser data, generate interval value
	movf    Stack1, W       ;              
	movwf   Minutes
	movf    Stack0, W
	movwf   Seconds
                          
	return 
;--------------------------------------
GetGPSByte
 if NMEAPol == 1        ; 1  =  RS232 levels (0/-ve = Idle
    btfss   GPSIN       ; 0  =  Logic levels (1 = Idle)
 else
    btfsc   GPSIN
 endif
    goto    GetGPSByte     ;Loop waiting for start bit

    call    NMEAHalfBitDel    ;Centre of start bit
 if NMEAPol == 1
    btfss   GPSIN
 else
    btfsc   GPSIN       ;Check it's still low
 endif
    goto    GetGPSByte

	movlw   8    
	movwf   Counter
    clrf    Temp        ;probably not needed, but good practice

InitByteLoop
    call    NMEAHalfBitDel
    call    NMEAHalfBitDel    ;
	nop
	bcf     STATUS , C      ;start in known state

 if NMEAPol == 1
    btfss   GPSIN
 else
    btfsc   GPSIN
 endif
	bsf     STATUS , C      ;Data into Carry
    rrf     Temp        ;and then into Temp, LSB first
	decfsz  Counter
    goto    InitByteLoop    ;loop is 8 + 2 * NMEAHalfBitDel long
	 
    call    NMEAHalfBitDel    ;#### Don't really need to wait for stop bit! ####
							; Now have approx 50us of stop bit left
    movf    Temp , W    ;to use this data

    return          ;Exit with received byte in W and Temp
;-----------------------------------------


;-----------------------------------------
NMEAHalfBitDel     ;Total delay 3.N + 5 clocks
	movlw   GPSBAUDDEL   
	movwf   DelCount
BaudDelLoop
	decfsz  DelCount
	goto    BaudDelLoop
	return
;-------------------------------
PushStack
	movf    StackE , W
	movwf   StackF
	movf    StackD , W  
	movwf   StackE
	movf    StackC , W
	movwf   StackD
	movf    StackB , W
	movwf   StackC
	movf    StackA , W
	movwf   StackB
	movf    Stack9 , W
	movwf   StackA
	movf    Stack8 , W
	movwf   Stack9
	movf    Stack7 , W
	movwf   Stack8
	movf    Stack6 , W
	movwf   Stack7 
	movf    Stack5 , W
	movwf   Stack6
	movf    Stack4 , W
	movwf   Stack5
	movf    Stack3 , W
	movwf   Stack4
	movf    Stack2 , W
	movwf   Stack3
	movf    Stack1 , W
	movwf   Stack2
	movf    Stack0 , W
	movwf   Stack1

	return
;-------------------------------
PNSeq       
    movlw   LOW(DutyCycle)
    call    GetEE       ;returns a value from 1 to 5 
    andlw   0x7F	;Remove hopping flag stored in MSB
    addlw   0xFF        ;Subtract 1

    btfsc   STATUS, Z
    return              ;For DCycle = 1, always return with W = 0 to Tx

    addlw   0xFF        ;Subtract 1 to give 0 - 3 for 50/33/25/20% DC
    movwf   Temp     
    rlf     Temp
    rlf     Temp, W
    andlw   0x0C        ;Multiply by 4 and mask to 3 maximum
    movwf   Temp        ;Temp is now offset in table corresponding to DCycle

    incf    PNCounter   ;Preloaded with 0xFF at startup so we get 0 first

    movlw   d'30'
    subwf   PNCounter, W	;W = Counter - d'30'   If +ve / zero (C=1) clear it
    btfsc   STATUS, C
    clrf    PNCounter

    movf    PNCounter, W
    movwf   PNLookup    ;Use PN counter to find correct bit in table
    rrf     PNLookup
    rrf     PNLookup 
    rrf     PNLookup, W ;Divide by 8 to point to appropriate byte
    andlw   0x03        ;Mask to 0 - 3
    movwf   PNLookup

    movf    Temp, W     ;Recover Duty Cycle table offset, (0- 3) * 4
    addwf   PNLookup    ;  and add into the lookup offset for a value 0 - 15
    call    PNTable     ;Extract the appropriate set of 8 bits
    movwf   PNLookup
                        
    movf    PNCounter, W    ;Now find correct bit to use within the byte
    andlw   0x07
    addlw   1
    movwf   Temp        ;Value 1 - 8 corresponding to bit to use
PNLoop
    rlf     PNLookup
    decfsz  Temp
    goto    PNLoop
                        ;Wanted bit is now in carry
    movlw   0
    btfss   STATUS, C
    movlw   1           ;0 in W transmits, 1 

    return
;----------------------
SendCWMsg           ;Enter with EE address programmed into CWMsgCnt
	movf    CWMsgCnt, W
	call    GetEE       ;Test for terminator at start = no CW message
	addlw   0
	btfsc   STATUS, Z
	return
	movlw   0           ;Lead in gap makes initial CW clearer
	call    OnOff
	call    CWDelay
	call    CWDelay
	call    CWDelay
	call    CWDelay

CWLoop
	movf    CWMsgCnt, W
	call    GetEE
	addlw   0
	btfsc   STATUS, Z
	goto    DoneCWMsg

	call    SendCWChar
	incf    CWMsgCnt
	goto    CWLoop
DoneCWMsg
	call    CWDelay
	call    CWDelay
	call    CWDelay
	call    CWDelay
	return
;--------------------------
SendCWChar      ;Sends a single CW character
	addlw   0xE0     ;Subtract 32 (add 256 - 32) to point to table
	call    CWTable
	movwf   CWChar      ;Doesn't affect STATUS so...
	movf    CWChar      ;  check for invalid returned code here
	btfsc   STATUS , Z      
	goto    NoChar

	movlw   0x07    
	andwf   CWChar , W      ;Extract element count into W
	movwf   CWCharCnt     
    	movlw   6       
	subwf   CWCharCnt , W     ;W = Elcount - 6.  If +ve/0 then C=1, > 5 els
	btfss   STATUS , C
	goto    Send5
	
	rrf     CWCharCnt   ;Puts LSB into C as 110 = .  111 = -
	call    SendEle     ;First  . or -  for 6 element characters
	movlw   5
	movwf   CWCharCnt     ;Force count to 5 for rest of the elements
Send5       
	rlf     CWChar      ;Element to be sent into C.
	call    SendEle     ;   count already in Counter
	decfsz  CWCharCnt   
	goto    Send5
NoChar          ;Inter word gap, and invalid codes.
	bcf     LEDRed
	call    CWDelay
	call    CWDelay
	return
;-------------------
SendEle         ;Data in STATUS , C;  0 = dit  1 = dah
	rrf     Stack0      ;Save C bit temporarily on stack
	movlw   1
    	call    OnOff
	bsf     LEDRed
	rlf     Stack0      ;Recovery C bit again, to get dot/dash 
	btfss   STATUS , C      ;test for dot or dash
	goto    SendDot
	call    CWDelay
	call    CWDelay
SendDot
	call    CWDelay
	movlw   0
    	call    OnOff
	bcf     LEDRed
    	call    CWDelay     ;Inter element gap
	return
;-------------------
CWDelay       
	movlw   CWSPEED
	movwf   DelCount2
CWDelLoop
	call    Delay1ms      
	decfsz  DelCount2
	goto    CWDelLoop
	return
;--------------------------
UpCase
    return

;-------------------------
DecodeHex   ;Single ASCII character to binary with error checking
        ;Untrapped characters remaining ...........
        ;at the moment Upper Case letters G-Z plus characters
        ;  : ; < = > ? @ [ \ ] ^ _ `    will be let through.
        ;Chars 0 - $2F and letters "g" up
        ;are detected.

        bsf     HexFlag         ;Flag for valid hex character
        movwf   Temp            ;
        movlw   0x30
        subwf   Temp            ;subtract $30 and leave in Temp
        btfss   STATUS , C      ;if < $30, result -ve so C=0, error, set=0
        bcf     HexFlag
        
        movf    Temp , W        ;Test for residual from above > d54 ie > "g"
        sublw   d'54'           ;W = 54 - W, if < 54  invalid char, -ve, C=0
        btfss   STATUS , C      ;
        bcf     HexFlag

        movf    Temp , W
        andlw   0x1F            ;mask to 5 bits, resolves upper / lower case
        movwf   Temp            ;if 0 - 9 then Temp is correct, else 7 high

        movf    Temp , W        
        sublw   d'9'            ;W = 9 - W
        btfsc   STATUS , C      ;if -ve, C=0, char > 9, so subtract 7
        goto    ConvDone
        movlw   d'7'
        subwf   Temp            ;
ConvDone
        movf    Temp , W        ;
        andlw   0x0F            ;mask out any rubbish left in
        return                  ;flag set if partly valid 
;-------------------------
InitGPS                 
	bsf	GPSOUT		;Send on off to A2200 module
	call	Delay1ms
	bcf	GPSOUT
	call	Delay1ms
	bsf	GPSOUT		;
	call	Delay1ms
	bcf	GPSOUT

	return
;-----------------------------------------
SpreadLoop
	movf	V1, W		;Chop Suey derived spreading waveform
	addwf	V2		;  something like 2^31 repeat rate ?????
	movf	V2, W
	addwf	V3
	movf	V3, W
	addwf	V4
	swapf	V4, W
	addwf	V1		;Result in V1
	return
;-----------------------------------------------
TestStart		;Entered every second after GPS is read
	bcf	StartFlag

	movlw	7		;Test/Demo Mode
	subwf	ModeInUse, w
	btfsc	STATUS, Z
	bsf	StartFlag	;Start immediately

	movlw	2
	subwf	ModeInUse, W	;W = Mode - 2, if >=2 C=1
	btfsc	STATUS, C
	goto	TestMinutes	;If Mode >=2 

	movf	ModeInUse, W
	btfss	STATUS, Z
	goto	Try30s

	movlw	d'15'
	subwf	Seconds, W
	btfsc	STATUS, Z
	bsf	StartFlag

	movlw	d'45'
	subwf	Seconds, W
	btfsc	STATUS, Z
	bsf	StartFlag		
Try30s		
	movlw	d'0'
	subwf	Seconds, W
	btfsc	STATUS, Z
	bsf	StartFlag	

	movlw	d'30'
	subwf	Seconds, W
	btfsc	STATUS, Z
	bsf	StartFlag	

	return
;...................
TestMinutes
	movf	Seconds, W
	btfss	STATUS, Z
	return			;Only test when seconds = 0 

	movlw	4
	subwf	ModeInUse, W	;Mode >=4 are multiples of 5 minutes
	btfsc	STATUS, C
	goto	FiveMins

	movlw	3		;2 minutes
	subwf	ModeInUse, W
	btfss	STATUS, Z
	goto	OneMin

	btfsc	Minutes, 0
	return		 
OneMin
	bsf	StartFlag
	return
;..............
FiveMins
	movlw	5
	subwf	ModeInUse, W	;>= Mode 5 15 minutes
	btfsc	STATUS, C
	goto	Try15m

	movlw	d'5'
	subwf	Minutes, W
	btfsc	STATUS, Z
	bsf	StartFlag

	movlw	d'10'
	subwf	Minutes, W
	btfsc	STATUS, Z
	bsf	StartFlag

	movlw	d'20'
	subwf	Minutes, W
	btfsc	STATUS, Z
	bsf	StartFlag

	movlw	d'25'
	subwf	Minutes, W
	btfsc	STATUS, Z
	bsf	StartFlag

	movlw	d'35'
	subwf	Minutes, W
	btfsc	STATUS, Z
	bsf	StartFlag

	movlw	d'40'
	subwf	Minutes, W
	btfsc	STATUS, Z
	bsf	StartFlag

	movlw	d'50'
	subwf	Minutes, W
	btfsc	STATUS, Z
	bsf	StartFlag

	movlw	d'55'
	subwf	Minutes, W
	btfsc	STATUS, Z
	bsf	StartFlag
	
Try15m
	movlw	6
	subwf	ModeInUse, W	;>= Mode 5 15 minutes
	btfsc	STATUS, C
	goto	Try30m

	movlw	d'15'
	subwf	Minutes, W
	btfsc	STATUS, Z
	bsf	StartFlag	

	movlw	d'45'
	subwf	Minutes, W
	btfsc	STATUS, Z
	bsf	StartFlag	
Try30m
	movlw	d'30'
	subwf	Minutes, W
	btfsc	STATUS, Z
	bsf	StartFlag	

	movlw	0
	subwf	Minutes, W
	btfsc	STATUS, Z
	bsf	StartFlag	

	return
;--------------------------------------
ReadCR			;RC Network t = 1.4 to 14ms  on PORT A4 charged from Vcc
	bsf	STATUS, RP0
	bcf	TRISA, 4
	bcf	STATUS, RP0	;A4 as output to discharge the cap
	bcf	PORTA, 4

	movlw	d'10'
	call	ProgDelay	;Allow 40ms discharge time

	clrf	RCValue	
	
	bsf	STATUS, RP0
	bsf	TRISA, 4
	bcf	STATUS, RP0	;A4 as input to monitor the charging
CRLoop
	movlw	d'60'		;To give time delay > longest timeconstant / 256
	movwf	Counter
CRTimDel
	decfsz	Counter		;Delay ~ 3N / 2.5MHz, so about 72us
	goto	CRTimDel	;  gives max count if threshold is at t=CR of ~ 194

	incf	RCValue
	btfsc	STATUS, Z
	decf	RCValue		;Freeze at 255 if it overflows

	btfss	PORTA, 4	;Wait for A4 to go high
	goto	CRLoop		;Exit with RC somewhere range 1 - 255

	bsf	STATUS, RP0
	bcf	TRISA, 4
	bcf	STATUS, RP0	;Tidy-up with A4 as output
	bcf	PORTA, 4	; and set to discharge cap

;............		;Now test resulting count against different thresholds
	THRESH	=	d'40'
	THRESHINCR = 	d'25'
	movlw	THRESH		;W = Value - Threshold, if -Ve, C=0
	subwf	RCValue, W	; get out with value loaded into W. 
	movlw	0
	btfss	STATUS, C	
	return
				;Otherwise carry on and test against next threshold
	movlw	THRESH + THRESHINCR
	subwf	RCValue, W	
	movlw	1
	btfss	STATUS, C
	return

	movlw	THRESH + THRESHINCR * 2
	subwf	RCValue, W	
	movlw	2
	btfss	STATUS, C
	return

	movlw	THRESH + THRESHINCR * 3
	subwf	RCValue, W	
	movlw	3
	btfss	STATUS, C
	return

	movlw	THRESH + THRESHINCR * 4
	subwf	RCValue, W	
	movlw	4
	btfss	STATUS, C
	return

	movlw	THRESH + THRESHINCR * 5
	subwf	RCValue, W	
	movlw	5
	btfss	STATUS, C
	return

	movlw	THRESH + THRESHINCR * 6
	subwf	RCValue, W	
	movlw	6
	btfss	STATUS, C
	return

	movlw	7
	return


	return			
;-----------------------------------------------
ReadSettings			;At turn on, read RC and modify Dutucycle value in EE
	call	ReadCR		;1 - 5 set duty cycle, 6 ignore, 7 hopping on, 8 hopping off
	sublw	7		;So pot works in opposite direction
	addlw	1		;Gives value from 1 - 8
	movwf	Temp

	movlw   LOW(DutyCycle)	
	movwf	Whi		;In preparation for storing new value
	call    GetEE
	movwf	iTemp		;Current EE value contains old duty cycle and hop status
				;Temp has current value of CR , iTemp has contents of EE

	movlw	7		;7 = Freq hop on, set MSB of EE location
	subwf	Temp, W
	btfss	STATUS, Z
	goto	Trymode8
	bsf	iTemp, 7	;Hop flag on
	goto	SaveDC		
;...........
Trymode8
	movlw	8		;Freq hop off, clear MSB of EE location
	subwf	Temp, W
	btfss	STATUS, Z
	goto	TryMode6
	bcf	iTemp, 7	;Hop flag off
	goto	SaveDC
;...........
TryMode6	
	movlw	6		;Reserved. Do nothing
	subwf	Temp, W
	btfsc	STATUS, Z
	goto	SaveDC

				;1 - 5 set duty cycle in LSBs of EE location.
	movlw	0xF8
	andwf	iTemp		;Clear LSBs 
	movf	Temp, W		;Recover value read from CR, 1 - 5
	andlw	0x07		;Mask it just in case
	addwf	iTemp		;Save ready for EE stored

SaveDC
	movf	iTemp, W
	call	StoreEE		;Whi was set earlier
	call	ShowDC

			;Now wait for TESTLINK to be reset
ModeWaitLoop

	bsf	LEDRed		;Unique flash pattern and bleeps
	bcf	LEDGrn
	movlw	d'25'
	call	ProgDelay
	bcf	LEDRed
	movlw	d'25'
	call	ProgDelay
	bsf	LEDGrn
	movlw	d'25'
	call	ProgDelay
	bcf	LEDGrn
	movlw	d'25'
	call	ProgDelay

	btfss	TESTLINK	
	goto	ModeWaitLoop

	return
;----------------------------------
ShowDC
	movlw	LOW(DutyCycle)
	call	GetEE
	movwf	iTemp

	movlw	"D"
	call	SendCWChar

	movf	iTemp, W
	andlw	0x07
	addlw	"0"
	call	SendCWChar		;Send duty cycle value 1 - 6

	movlw	d'100'
	call	ProgDelay

	movlw	"X"
	btfsc	iTemp, 7
	movlw	"H"		;Send X for no hop, H for Hop mode
	call	SendCWChar
	
	movlw	d'200'
	call	ProgDelay

	movlw	"D"		;Send it again
	call	SendCWChar
	movf	iTemp, W
	andlw	0x07
	addlw	"0"
	call	SendCWChar		

	movlw	d'100'
	call	ProgDelay

	movlw	"X"
	btfsc	iTemp, 7
	movlw	"H"		
	call	SendCWChar

	return
;============================================
  org (($ >> 8) + 1) * 0x100		;Next page after end of code.   All tables in the same page
	
;-------------------
RampTable			;Raised Cosine amplitude ramp
	andlw	0x0F		;Just in case
	addwf   PCL	
	dt	0x00,0x02,0x0B,0x18,0x2A,0x3F,0x58,0x72,0x8D,0xA7,0xBF,0xD5,0xE7,0xF4,0xFD,0xFF
;---------------------------------------------------
PTable			;Gaussian shaped half pulse
		addwf   PCL
	dt	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00	;Access with input in range 0 - 64
	dt	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	dt	0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x02, 0x04
	dt	0x07, 0x0C, 0x14, 0x1E, 0x2B, 0x3C, 0x50, 0x66
	dt	0x7D, 0x94, 0xAA, 0xBE, 0xCF, 0xDC, 0xE6, 0xEE
	dt	0xF3, 0xF6, 0xF8, 0xF9, 0xF9, 0xFA, 0xFA, 0xFA
	dt	0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA
	dt	0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA, 0xFA 
;--------------------------
PNTable
    	movf    PNLookup, W
    	andlw   0x0F
	addwf   PCL

    	dt      b'10100101', b'10101000', b'11010110', b'10101000'   ;50%
    	dt      b'10010101', b'00101000', b'01000100', b'10001000'   ;33%
    	dt      b'10000100', b'00110010', b'00100100', b'00100000'   ;27%
    	dt      b'10000100', b'00000010', b'01000100', b'00010000'   ;20%
;---------------------------------------
GetRegLen       ;gets number of bytes needed for value in DDSAddr
	movf    DDSAddr, W
	addwf   PCL

	dt      2,2,6,6,6,4,3,4,2,2,1,2       ;Number of bytes in each register
;-------------------------
DataTo9852          ;Send contents of Dn...D0 to 9852   Has computed jumps so needs 
    	movf    DDSAddr, W  				; to be in the tables page
    	call    Send9852Byte    ;Send the address data

    	call    GetRegLen       ;Returns with number of bytes to be sent
	movwf   Temp
	bcf     STATUS, C
	rlf     Temp, W
    	sublw   d'12'           ;W = 12 - 2 * (No of bytes to be sent)
    	addwf   PCL         ;Jump into byte sending at the right place

	movf    D5, W
	call    Send9852Byte
	movf    D4, W
	call    Send9852Byte
	movf    D3, W
	call    Send9852Byte
	movf    D2, W
	call    Send9852Byte
	movf    D1, W
	call    Send9852Byte
	movf    D0, W
	call    Send9852Byte

    	bcf     SDIO        ;Rest in low state for min. current consumption
	return
;-------------------------
CWTable       ;CW Characters in compressed form. 
	addwf   PCL
	
	dt   b'00000000'  ; [sp] 00
	dt   b'00000000'  ;  !   00
	dt   b'10010110'  ;  "   96
	dt   b'00010101'  ;  #   15
	dt   b'00000000'  ;  $   00
	dt   b'00000000'  ;  %   00
	dt   b'00000000'  ;  &   00
	dt   b'11110110'  ;  '   F6
	dt   b'10110101'  ;  (   B5
	dt   b'01101111'  ;  )   6F
	dt   b'01000101'  ;  *   45
	dt   b'00000000'  ;  +   00
	dt   b'10011111'  ;  ,   9F     Coding for CW characters of up to
	dt   b'00001111'  ;  -   0F      6 elements.  Stored as :
	dt   b'10101110'  ;  .   AE     Bits 7 to 3 make first five CW elements
	dt   b'10010101'  ;  /   95     Reading  Left to Right 
	dt   b'11111101'  ;  0   FD     Coded as 0 = dit   1 = dah
	dt   b'01111101'  ;  1   7D     Bts 2-0 are number of elements except for 
	dt   b'00111101'  ;  2   3D     110 > 1st element is a dit  
	dt   b'00011101'  ;  3   1D     111 > 1st element is a dah       
	dt   b'00001101'  ;  4   0D
	dt   b'00000101'  ;  5   05
	dt   b'10000101'  ;  6   85
	dt   b'11000101'  ;  7   C5
	dt   b'11100101'  ;  8   E5
	dt   b'11110101'  ;  9   F5
	dt   b'00000000'  ;  :   00
	dt   b'00000000'  ;  ;   00
	dt   b'00000000'  ;  <   00
	dt   b'10001101'  ;  =   8D
	dt   b'00000000'  ;  >   00
	dt   b'01100110'  ;  ?   66
	dt   b'11010110'  ;  @   D6
	dt   b'01000010'  ;  A   42
	dt   b'10000100'  ;  B   84
	dt   b'10100100'  ;  C   A4
	dt   b'10000011'  ;  D   83
	dt   b'00000001'  ;  E   01
	dt   b'00100100'  ;  F   24
	dt   b'11000011'  ;  G   C3
	dt   b'00000100'  ;  H   04
	dt   b'00000010'  ;  I   02
	dt   b'01110100'  ;  J   74
	dt   b'10100011'  ;  K   A3
	dt   b'01000100'  ;  L   44
	dt   b'11000010'  ;  M   C2
	dt   b'10000010'  ;  N   82
	dt   b'11100011'  ;  O   E3
	dt   b'01100100'  ;  P   64
	dt   b'11010100'  ;  Q   D4
	dt   b'01000011'  ;  R   43
	dt   b'00000011'  ;  S   03
	dt   b'10000001'  ;  T   81
	dt   b'00100011'  ;  U   23
	dt   b'00010100'  ;  V   14
	dt   b'01100011'  ;  W   63
	dt   b'10010100'  ;  X   94
	dt   b'10110100'  ;  Y   B4
	dt   b'11000100'  ;  Z   C4
;-----------------------------------------------------
TimeTable		;Cycle time in units of 15s
	dt	1,2,4,d'8',d'20',d'60',d'120'
	addwf	PCL
;--------------------------
ModeDivTable
	addwf	PCL
; Interrupt Divider value = d'65536' - MODEDIV + d'2'	;Prescale / 8,  Int overhead = 16
	dt	HIGH(d'65538' - MODEDIV15) , LOW(d'65538' - MODEDIV15)
	dt	HIGH(d'65538' - MODEDIV30) , LOW(d'65538' - MODEDIV30)
	dt	HIGH(d'65538' - MODEDIV60) , LOW(d'65538' - MODEDIV60)
	dt	HIGH(d'65538' - MODEDIV120) , LOW(d'65538' - MODEDIV120)
	dt	HIGH(d'65538' - MODEDIV300) , LOW(d'65538' - MODEDIV300)
	dt	HIGH(d'65538' - MODEDIV900) , LOW(d'65538' - MODEDIV900)
	dt	HIGH(d'65538' - MODEDIV1800) , LOW(d'65538' - MODEDIV1800)
	dt	HIGH(d'65538' - MODEDIVTEST) , LOW(d'65538' - MODEDIVTEST)


 ;  Table space nearly used up

;-----------------------------------------------------
 org (($ >> 8) + 1) * 0x100		;Next page after end of last table
	de	"FST4BCN_9852"

;================================
 
;--------------------------------
    cblock 0x20
        Temp, iTemp
        DelCount, DelCount2
        Counter
        SymbCount       ;Counts 0 to max for each FST4 symbol
        PosData         ;Position of the 2 data bits in Byte from EE
        SoftCount       ;Used in amplitude ramping
        Amplitude
        DDSAddr
        Flags
        Wreg, Whi
        SaveSts
        Seconds, Minutes
        ICount
        CommaCount
        CWChar
        CWMsgCnt,  CWCharCnt
        RxChar
        TonePtr
        SerCounter      ;Count of characters on GPS serial data
        RxStatus    	;GPS Receiver status
	BCDHi, BCDLo
        D0, D1, D2, D3, D4, D5	;D data sent to DDS registers
        A0, A1, A2, A3, A4
        B0, B1, B2, B3, B4  ;Qx output from maths routines
        Q0, Q1, Q2, Q3, Q4
	V1, V2, V3, V4		;Randomising
	RandOffset
	QuadSymbol, LastSymbol, ThisSymbol, NextSymbol, SampCount
        PNLookup, PNCounter, CheckDigit
	IptCountHi, IptCountLo
	ModeInUse
	RCValue
        Stack0, Stack1, Stack2, Stack3, Stack4, Stack5, Stack6, Stack7
        Stack8, Stack9, StackA, StackB, StackC, StackD, StackE, StackF
    endc   
;==========================================
		end

;Mode Tone Spacing (Hz)	Sample Int (s)	Tx period (160 symbols, s)
;-15   	16.6667		0.06		9.6
;-30	7.14		0.14		22.4
;-60	3.09		0.32375		51.8
;-120	1.46		0.68313		109.3
;-300	0.558		1.7919		286.7
;-900	0.180		5.547		887.5
;-1800	0.089		11.2		1792
;
